# coding=utf-8
"""Folders / Documents module."""
import typing

if typing.TYPE_CHECKING:
    from abilian.sbe.app import Application


def register_plugin(app: "Application") -> None:
    from .views import blueprint
    from .models import setup_listener
    from .cli import antivirus
    from . import signals  # noqa
    from . import lock

    app.register_blueprint(blueprint)
    setup_listener()

    # set default lock lifetime
    app.config.setdefault("SBE_LOCK_LIFETIME", lock.DEFAULT_LIFETIME)

    app.cli.add_command(antivirus)



# coding=utf-8
"""Celery tasks related to document transformation and preview."""
import logging
import typing
from contextlib import contextmanager
from typing import Any, Iterator, Optional, Tuple, Union

from abilian.core.extensions import db
from abilian.services import converter, get_service
from abilian.services.conversion import ConversionError, HandlerNotFound
from celery import shared_task
from sqlalchemy.orm import Session

if typing.TYPE_CHECKING:
    from .models import Document

logger = logging.getLogger(__package__)


@contextmanager
def get_document(
    document_id: int, session: Optional[Session] = None
) -> Iterator[Tuple[Session, Optional["Document"]]]:
    """Context manager that yields (session, document)."""
    from .models import Document

    if session is None:
        doc_session = db.create_scoped_session()
    else:
        doc_session = session

    with doc_session.begin_nested():
        query = doc_session.query(Document)
        document = query.get(document_id)
        yield (doc_session, document)

    # cleanup
    if session is None:
        doc_session.commit()
        doc_session.close()


@shared_task
def process_document(document_id: int) -> None:
    """Run document processing chain."""
    with get_document(document_id) as (session, document):
        if document is None:
            return

        # True = Ok, None means no check performed (no antivirus present)
        is_clean = _run_antivirus(document)
        if is_clean is False:
            return

    preview_document.delay(document_id)
    convert_document_content.delay(document_id)


def _run_antivirus(document: "Document") -> Optional[bool]:
    antivirus = get_service("antivirus")
    if antivirus and antivirus.running:
        is_clean = antivirus.scan(document.content_blob)
        if "antivirus_task" in document.content_blob.meta:
            del document.content_blob.meta["antivirus_task"]
        return is_clean
    return None


@shared_task
def antivirus_scan(document_id):
    """Return antivirus.scan() result."""
    with get_document(document_id) as (session, document):
        if document is None:
            return
        return _run_antivirus(document)


@shared_task
def preview_document(document_id: int) -> None:
    """Compute the document preview images with its default preview size."""
    with get_document(document_id) as (session, document):
        if document is None:
            # deleted after task queued, but before task run
            return

        try:
            converter.to_image(
                document.content_digest,
                document.content,
                document.content_type,
                0,
                document.preview_size,
            )
        except ConversionError as e:
            logger.info(
                "Preview failed: %s", str(e), exc_info=True, extra={"stack": True}
            )


@shared_task
def convert_document_content(document_id: int) -> None:
    """Convert document content."""
    with get_document(document_id) as (session, doc):
        if doc is None:
            # deleted after task queued, but before task run
            return

        convert_to_pdf(doc)
        convert_to_text(doc)
        extract_metadata(doc)


def convert_to_pdf(doc: "Document") -> None:
    error_kwargs = {"exc_info": True, "extra": {"stack": True}}

    if doc.content_type == "application/pdf":
        doc.pdf = doc.content
    else:
        try:
            doc.pdf = converter.to_pdf(
                doc.content_digest, doc.content, doc.content_type
            )
        except HandlerNotFound:
            doc.pdf = b""
        except ConversionError as e:
            doc.pdf = b""
            logger.info(
                "Conversion to PDF failed for document %s: %s",
                doc.name,
                e,
                **error_kwargs
            )


def convert_to_text(doc: "Document") -> None:
    error_kwargs = {"exc_info": True, "extra": {"stack": True}}

    try:
        doc.text = converter.to_text(doc.content_digest, doc.content, doc.content_type)
    except ConversionError as e:
        doc.text = ""
        logger.info(
            "Conversion to text failed for document %s: %s", doc.name, e, **error_kwargs
        )


def extract_metadata(doc: "Document") -> None:
    error_kwargs = {"exc_info": True, "extra": {"stack": True}}

    doc.extra_metadata = {}
    try:
        doc.extra_metadata = converter.get_metadata(
            doc.content_digest, doc.content, doc.content_type
        )
    except ConversionError as e:
        logger.warning(
            "Metadata extraction failed on document %s: %s", doc.name, e, **error_kwargs
        )
    except UnicodeDecodeError as e:
        logger.error("Unicode issue on document %s: %s", doc.name, e, **error_kwargs)
    except Exception as e:
        logger.error("Other issue on document %s: %s", doc.name, e, **error_kwargs)

    if doc.text:
        import langid

        doc.language = langid.classify(doc.text)[0]

    doc.page_num = doc.extra_metadata.get("PDF:Pages", 1)



# coding=utf-8

from typing import Any

import pytest
from pytest import fixture
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session

from abilian.sbe.app import Application
from abilian.sbe.apps.documents.models import Document, Folder
from abilian.sbe.apps.documents.repository import Repository


@fixture
def repository(app: Application) -> Repository:
    repository = Repository()
    repository.init_app(app)
    return repository


@fixture
def root(session: Session) -> Folder:
    root = Folder(title="")
    session.add(root)
    session.flush()
    return root


def test_create_doc(root: Folder, session: Session) -> None:
    doc = root.create_document("doc")
    session.flush()
    assert doc.path == "/doc"
    assert len(root.children) == 1
    assert not doc.is_root_folder


def test_create_folder(root: Folder, session: Session) -> None:
    folder = root.create_subfolder("folder")
    session.flush()
    assert folder.title == "folder"
    assert folder.name == "folder"
    assert folder.path == "/folder"
    assert len(root.children) == 1
    assert len(folder.children) == 0
    assert not folder.is_root_folder


def test_move(root: Folder, repository: Repository) -> None:
    doc = root.create_document("doc")
    folder = root.create_subfolder("folder")

    repository.move_object(doc, folder, "newdoc")
    assert doc in folder.children
    assert doc not in root.children
    assert len(root.children) == 1
    assert len(folder.children) == 1
    assert doc.title == "newdoc"
    assert doc.path == "/folder/newdoc"


def test_copy(root: Folder, repository: Repository) -> None:
    doc = root.create_document("doc")
    folder = root.create_subfolder("folder")

    doc_copy = repository.copy_object(doc, folder, "copydoc")

    assert len(root.children) == 2
    assert len(folder.children) == 1
    assert doc_copy.title == "copydoc"
    assert doc_copy.name == "copydoc"
    assert doc_copy.path == "/folder/copydoc"
    assert doc_copy in folder.children


def test_copy_nested_folders(root: Folder, repository: Repository) -> None:
    folder1 = root.create_subfolder("folder1")
    folder2 = root.create_subfolder("folder2")
    subfolder = folder1.create_subfolder("subfolder")

    folder1_copy = repository.copy_object(folder1, folder2)

    assert len(root.children) == 2
    assert folder1 in root.children
    assert folder2 in root.children

    assert len(folder1.children) == 1
    assert subfolder in folder1.children

    assert len(folder2.children) == 1
    assert folder1_copy in folder2.children


def test_move_nested_folders(root: Folder, repository: Repository) -> None:
    folder1 = root.create_subfolder("folder1")
    folder2 = root.create_subfolder("folder2")
    subfolder = folder1.create_subfolder("subfolder")  # noqa

    repository.move_object(folder1, folder2)

    assert len(root.children) == 1
    assert len(folder1.children) == 1


def test_rename(root: Folder, repository: Repository) -> None:
    folder = root.create_subfolder("folder")
    doc = folder.create_document("doc")

    repository.rename_object(doc, "doc1")
    assert doc.title == "doc1"
    assert doc.name == "doc1"
    assert doc.path == "/folder/doc1"

    repository.rename_object(folder, "folder1")
    assert folder.title == "folder1"
    assert folder.name == "folder1"
    assert folder.path == "/folder1"
    assert doc.path == "/folder1/doc1"


def test_delete(root: Folder, repository: Repository, session: Session) -> None:
    doc = root.create_document("doc")
    folder = root.create_subfolder("folder")
    session.flush()

    assert doc.parent == root
    assert folder.parent == root
    assert doc in root.children
    assert folder in root.children
    assert not doc.is_root_folder
    assert not folder.is_root_folder

    doc_id = doc.id
    folder_id = folder.id

    repository.delete_object(doc)
    repository.delete_object(folder)
    assert len(root.children) == 0

    session.flush()
    assert Folder.query.get(folder_id) is None
    assert Document.query.get(doc_id) is None

    # test delete tree
    folder = root.create_subfolder("folder")
    sub = folder.create_subfolder("subfolder")  # noqa
    doc = folder.create_document("doc")  # noqa
    session.flush()

    repository.delete_object(folder)
    assert len(root.children) == 0
    assert Folder.query.all() == [root]
    assert Document.query.all() == []


def test_no_duplicate_name(root: Folder, session: Session) -> None:
    root.create_subfolder("folder_1")
    root.create_subfolder("folder_1")
    with pytest.raises(IntegrityError):
        session.flush()



# coding=utf-8
""""""
from datetime import datetime, timedelta
from unittest import mock

from abilian.testing.util import client_login
from flask import url_for
from flask_login import login_user

from abilian.sbe.apps.communities.models import MANAGER, MEMBER
from abilian.sbe.apps.forum.tests.util import get_string_from_file

from ..cli import inject_email
from ..models import Post, Thread
from ..tasks import build_reply_email_address, extract_email_destination


def test_posts_ordering(db, community1):
    thread = Thread(community=community1, title="test ordering")
    db.session.add(thread)
    t1 = datetime(2014, 6, 20, 15, 0, 0)
    p1 = Post(thread=thread, body_html="post 1", created_at=t1)
    t2 = datetime(2014, 6, 20, 15, 1, 0)
    p2 = Post(thread=thread, body_html="post 2", created_at=t2)
    db.session.flush()
    p1_id, p2_id = p1.id, p2.id
    assert [p.id for p in thread.posts] == [p1_id, p2_id]

    # set post1 created after post2
    t1 = t1 + timedelta(minutes=2)
    p1.created_at = t1
    db.session.flush()
    db.session.expire(thread)  # force thread.posts refreshed from DB
    assert [p.id for p in thread.posts] == [p2_id, p1_id]


def test_thread_indexed(app, db, community1, community2, req_ctx):
    index_svc = app.services["indexing"]
    index_svc.start()
    security_svc = app.services["security"]
    security_svc.start()

    thread1 = Thread(title="Community 1", community=community1)
    db.session.add(thread1)

    thread2 = Thread(title="Community 2: other", community=community2)
    db.session.add(thread2)
    db.session.commit()

    index_svc = app.services["indexing"]
    obj_types = (Thread.entity_type,)

    login_user(community1.test_user)
    res = index_svc.search("community", object_types=obj_types)
    assert len(res) == 1
    hit = res[0]
    assert hit["object_key"] == thread1.object_key

    login_user(community2.test_user)
    res = index_svc.search("community", object_types=obj_types)
    assert len(res) == 1
    hit = res[0]
    assert hit["object_key"] == thread2.object_key


def test_forum_home(client, community1, login_admin, req_ctx):
    response = client.get(url_for("forum.index", community_id=community1.slug))
    assert response.status_code == 200


def test_create_thread_informative(app, db, client, community1, req_ctx):
    """Test with 'informative' community.

    No mail sent, unless user is MANAGER
    """
    user = community1.test_user
    assert community1.type == "informative"
    community1.set_membership(user, MEMBER)
    db.session.commit()

    title = "Brand new thread"
    content = "shiny thread message"
    url = url_for("forum.new_thread", community_id=community1.slug)
    data = {"title": title, "message": content}
    data["__action"] = "create"

    mail = app.extensions["mail"]
    with client_login(client, user):
        with mail.record_messages() as outbox:
            data["send_by_email"] = "y"  # actually should not be in html form
            response = client.post(url, data=data)
            assert response.status_code == 302
            # FIXME: this doesn't pass
            # assert len(outbox) == 0

        community1.set_membership(user, MANAGER)
        db.session.commit()

        with mail.record_messages() as outbox:
            data["send_by_email"] = "y"  # should be in html form
            response = client.post(url, data=data)
            assert response.status_code == 302
            assert len(outbox) == 1

        with mail.record_messages() as outbox:
            del data["send_by_email"]
            response = client.post(url, data=data)
            assert response.status_code == 302
            assert len(outbox) == 0


def test_build_reply_email_address(app, req_ctx):
    post = mock.Mock()
    post.id = 2
    post.thread_id = 3

    member = mock.Mock()
    member.id = 4

    result = build_reply_email_address("test", post, member, "example.com")
    expected = "test+P-en-3-4-a8f33983311589176c711111dc38d94d@example.com"
    assert result == expected


def test_extract_mail_destination(app, req_ctx):
    # app.config['MAIL_SENDER'] = 'test@testcase.app.tld'

    test_address = "test+P-en-3-4-a8f33983311589176c711111dc38d94d@example.com"
    infos = extract_email_destination(test_address)
    assert infos == ("en", "3", "4")

    test_address = (
        "John Q Public <test+P-en-3-4-a8f33983311589176c711111dc38d94d@example.com>"
    )
    infos = extract_email_destination(test_address)
    assert infos == ("en", "3", "4")

    test_address = (
        '"John Q Public" <test+P-en-3-4-a8f33983311589176c711111dc38d94d@example.com>'
    )
    infos = extract_email_destination(test_address)
    assert infos == ("en", "3", "4")


def test_create_thread_and_post(community1, client, app, db, req_ctx):
    community = community1
    user = community.test_user

    # activate email reply
    app.config["SBE_FORUM_REPLY_BY_MAIL"] = True

    # create a new user, add him/her to the current community as a MANAGER
    community.set_membership(user, MANAGER)
    db.session.commit()
    client_login(client, user)

    mail = app.extensions["mail"]
    with mail.record_messages() as outbox:
        title = "Brand new thread"
        content = "shiny thread message"
        url = url_for("forum.new_thread", community_id=community.slug)
        data = {"title": title, "message": content}
        data["__action"] = "create"
        data["send_by_email"] = "y"
        response = client.post(url, data=data)
        assert response.status_code == 302

        # extract the thread_id from the redirection url in response
        threadid = response.location.rsplit("/", 2)[1]

        # retrieve the new thread, make sur it has the message
        url = url_for(
            "forum.thread", thread_id=threadid, community_id=community.slug, title=title
        )
        response = client.get(url)
        assert response.status_code == 200
        assert content in response.get_data(as_text=True)

        # check the email was sent with the new thread
        assert len(outbox) == 1
        assert outbox[0].subject == "[My Community] Brand new thread"

    # reset the outbox for checking threadpost email
    with mail.record_messages() as outbox:
        content = data["message"] = "my cherished post"
        del data["title"]
        response = client.post(url, data=data)
        assert response.status_code == 302

        # retrieve the new thread, make sur it has the message
        url = url_for(
            "forum.thread", thread_id=threadid, community_id=community.slug, title=title
        )
        response = client.get(url)
        assert response.status_code == 200
        assert content in response.get_data(as_text=True)

        # check the email was sent with the new threadpost
        assert len(outbox) == 1
        expected = "[My Community] Brand new thread"
        assert str(outbox[0].subject) == expected


@mock.patch("fileinput.input")
@mock.patch("abilian.sbe.apps.forum.commands.process_email")
def test_parse_forum_email(mock_process_email, mock_email):
    """No processing is tested only parsing into a email.message and
    verifying inject_email() logic."""
    # first load a test email returned by the mock_email
    mock_email.return_value = get_string_from_file("notification.email")

    # test the parsing function
    inject_email()

    # assert the email is read
    assert mock_email.called
    # assert a call on the celery task was made implying a message creation
    assert mock_process_email.delay.called

    ##
    mock_email.reset_mock()
    mock_process_email.delay.reset_mock()
    assert not mock_email.called
    assert not mock_process_email.delay.called

    mock_email.return_value = get_string_from_file("defects.email")
    inject_email()
    assert mock_email.called
    assert not mock_process_email.delay.called



# coding=utf-8
from email.feedparser import FeedParser
from pathlib import Path


def get_string_from_file(filename="notification.email"):
    """Load a test email, return as string."""
    filepath = Path(__file__).parent / "data" / filename
    with filepath.open("rt", encoding="utf-8") as email_file:
        email_string = email_file.read()
    return email_string


def get_email_message_from_file(filename="notification.email"):
    """Load a mail and parse it into a email.message."""
    email_string = get_string_from_file(filename)
    parser = FeedParser()
    parser.feed(email_string)
    email_message = parser.close()
    return email_message



# coding=utf-8
"""Forum views."""

from datetime import datetime
from typing import List, Union

from abilian.i18n import _l
from abilian.services.viewtracker import viewtracker
from flask import g
from flask_babel import format_date

from abilian.sbe.apps.communities.security import is_manager
from abilian.sbe.apps.documents.models import Document
from abilian.sbe.apps.wiki.models import WikiPage


def object_viewers(entity: Union[Document, WikiPage]) -> List:
    if is_manager():
        views = viewtracker.get_views(entity=entity)
        community_members_id = [
            user.id for user in g.community.members if user.id != entity.creator.id
        ]
        viewers = []
        for view in views:
            if view.user_id in set(community_members_id):
                viewers.append(
                    {"user": view.user, "viewed_at": view.hits[-1].viewed_at}
                )
        return viewers
    return []


def activity_time_format(time: datetime, now: datetime = None) -> str:
    if not time:
        return ""

    if not now:
        now = datetime.utcnow()
    time_delta = now - time
    month_abbreviation = format_date(time, "MMM")
    days, hours, minutes, seconds = (
        time_delta.days,
        time_delta.seconds // 3600,
        time_delta.seconds // 60,
        time_delta.seconds,
    )

    if days == 0 and hours == 0 and minutes == 0:
        return "{}{}".format(seconds, _l("s"))

    if days == 0 and hours == 0:
        return "{}{}".format(minutes, _l("m"))

    if days == 0:
        return "{}{}".format(hours, _l("h"))

    if days < 30:
        return "{}{}".format(days, _l("d"))

    if time.year == now.year:
        return f"{month_abbreviation} {time.day}"

    return "{} {}".format(month_abbreviation, str(time.year))



# coding=utf-8
"""Static configuration for the application.

TODO: add more (runtime) flexibility in plugin discovery, selection
and activation.
"""
import logging

import jinja2
from abilian.app import Application as BaseApplication
from abilian.core.celery import FlaskCelery as BaseCelery
from abilian.core.celery import FlaskLoader as CeleryBaseLoader
from abilian.services import converter

from .apps.documents.repository import repository
from .extension import sbe

# Used for side effects, do not remove

__all__ = ["create_app", "Application"]

logger = logging.getLogger(__name__)


def create_app(config=None, **kw):
    app = Application(**kw)
    app.setup(config)
    return app


# loader to be used by celery workers
class CeleryLoader(CeleryBaseLoader):
    flask_app_factory = "abilian.sbe.app.create_app"


class Application(BaseApplication):

    APP_PLUGINS = BaseApplication.APP_PLUGINS + (
        "abilian.sbe.apps.main",
        "abilian.sbe.apps.notifications",
        "abilian.sbe.apps.preferences",
        "abilian.sbe.apps.wiki",
        "abilian.sbe.apps.wall",
        "abilian.sbe.apps.documents",
        "abilian.sbe.apps.forum",
        # "abilian.sbe.apps.calendar",
        "abilian.sbe.apps.communities",
        "abilian.sbe.apps.social",
        "abilian.sbe.apps.preferences",
    )

    def setup(self, config):
        super().setup(config)
        loader = jinja2.PackageLoader("abilian.sbe", "templates")
        self.register_jinja_loaders(loader)

    def init_extensions(self):
        BaseApplication.init_extensions(self)
        sbe.init_app(self)
        repository.init_app(self)
        converter.init_app(self)
